import BookTrackerPlugin from "@src/main";
import { App, normalizePath } from "obsidian";
import { EventEmitter } from "./event";

interface StorageEventMap {
	change: { path: string };
}

export class Storage extends EventEmitter<StorageEventMap> {
	private readonly app: App = this.plugin.app;
	private readonly baseDir = this.plugin.manifest.dir!;

	public constructor(private readonly plugin: BookTrackerPlugin) {
		super();
		plugin.registerEvent(
			// @ts-expect-error "raw" event is an internal api
			this.app.vault.on("raw", this.fileChangeHandler.bind(this))
		);
	}

	private getFilePath(filename: string): string {
		return normalizePath(`${this.baseDir}/${filename}`.replace(/\/$/, ""));
	}

	public async readJSON<T>(filename: string): Promise<T | null> {
		const filePath = this.getFilePath(filename);
		const content = await this.app.vault.adapter.read(filePath);
		if (!content) {
			return null;
		}

		try {
			return JSON.parse(content) as T;
		} catch (error) {
			console.error(`Error parsing JSON from ${filePath}:`, error);
			return null;
		}
	}

	public async writeJSON<T>(filename: string, data: T): Promise<void> {
		const filePath = this.getFilePath(filename);
		const content = JSON.stringify(data, null, 2);
		await this.app.vault.adapter.write(filePath, content);
	}

	public async listFiles(subdir = ""): Promise<string[]> {
		const files = await this.app.vault.adapter.list(
			this.getFilePath(subdir)
		);
		return files.files;
	}

	public async listFolders(): Promise<string[]> {
		const files = await this.app.vault.adapter.list(this.baseDir);
		return files.folders;
	}

	private fileChangeHandler(path: string) {
		if (!path.startsWith(this.baseDir)) return;

		path = path.replace(this.baseDir + "/", "");

		this.emit("change", { path });
	}
}
